/**
 * Buffer对象的 polyfill
 * @author yutent<yutent.io@gmail.com>
 * @date 2024/05/29 12:29:17
 */

import {
  encode,
  decode,
  getType,
  to16,
  hex2u8,
  sum,
  compare
} from './helper.js'
import { base64encode, base64decode, str2bin, bin2str } from './base64.js'

/**
 * 将其他类型的值, 转为uint8array
 */
function any2u8(data = '', encoding = 'utf8') {
  if (typeof data === 'object') {
    let type = getType(data)

    switch (type) {
      case 'Buffer':
      case 'Uint8Array':
        return data

      case 'ArrayBuffer':
        break

      case 'DataView':
        data = data.buffer
        break
    }
    return new Uint8Array(data)
  } else {
    switch (encoding.toLowerCase()) {
      case 'base64':
        data = base64decode(data)
        break

      case 'hex':
        return hex2u8(data)

      case 'binary':
        data = bin2str(data)
        break
    }

    return encode(data + '')
  }
}

/**
 * 创建基于Uint8Array的Buffer类
 */
export default class Buffer extends Uint8Array {
  get [Symbol.toStringTag]() {
    return 'Buffer'
  }

  /**
   * 返回指定内容的字节数
   */
  static byteLength(data, encoding = 'utf8') {
    let u8 = any2u8(data, encoding)
    return u8.length
  }

  /**
   * 从给定的内容中创建Buffer对象
   * @param data <any>
   * @return <Buffer>
   */
  static from(data, offsetOrEncoding, end) {
    let buf = data
    let start = 0
    let _encoding = 'utf8'

    if (offsetOrEncoding !== void 0) {
      if (typeof offsetOrEncoding === 'number') {
        start = offsetOrEncoding
      } else if (typeof offsetOrEncoding === 'string') {
        _encoding = offsetOrEncoding
      }
    }

    buf = any2u8(data, _encoding)

    return new Buffer(buf.slice(start, end))
  }

  /**
   * 从给定的内容中创建Buffer对象
   * @param len <Int> Buffer长度
   * @param str <String> 创建Buffer时, 填充的内容
   * @return <Buffer>
   */
  static alloc(len, str = '', encoding) {
    let buf = new Buffer(len)
    if (str) {
      buf.fill(str, 0, len, encoding)
    }
    return buf
  }

  /**
   * 合并多个Buffer对象
   */
  static concat(list, totalLength) {
    let len = totalLength || sum(list)
    let buf = Buffer.alloc(len)

    let offset = 0
    for (let it of list) {
      it.copy(buf, offset)
      offset += it.length
    }

    return buf
  }

  /**
   * 判断是否为Buffer对象
   */
  static isBuffer(target) {
    return getType(target) === 'Buffer'
  }

  /**
   * 比较2个对象
   */
  static compare = compare

  /**
   * 替换指定字节内容, fill, copy, write, 都会用到这个方法
   */
  #replace(target, start, end) {
    for (let i = 0; i < end - start; i++) {
      this[start + i] = target[i]
    }
  }

  toJSON() {
    return { type: 'Buffer', data: [...this].map(to16) }
  }

  /**
   * 这个重构为nodejs版的buf.toString()的功能
   */
  toString(encoding = 'utf8', start = 0, end) {
    let buf = this.slice(start, end)

    switch (encoding) {
      case 'base64':
        return base64encode(buf)

      case 'hex':
        return [...buf].map(to16).join('')

      case 'binary':
        return str2bin(decode(buf))

      default:
        return decode(buf)
    }
  }

  /**
   * 原生的 Uint8Array的fill只支持数字,
   * 这里拓展, 支持字符串和其他Array对象, 与nodejs版相似
   */
  fill(val, start = 0, end, encoding = 'utf8') {
    //
    end ??= this.length

    if (typeof val === 'number') {
      super.fill(val, start, end)
    } else {
      val = [...any2u8(val, encoding)]
      if (val.length === 0) {
        val = [0]
      }
      let tmp = val

      while (tmp.length < end - start) {
        tmp = tmp.concat(val)
      }
      tmp = tmp.slice(0, end)
      this.#replace(tmp, start, end)
    }
  }

  /**
   * 向目标对象中复制内容 (改变的是目标对象, 不是自己)
   */
  copy(target, targetStart = 0, sourceStart = 0, sourceEnd) {
    if (!(target instanceof Buffer)) {
      throw new TypeError(
        "copy()'s argument target must be a Buffer Object, but got %s",
        getType(target)
      )
    }
    sourceEnd ??= this.length
    // 复制的总长度, 默认等于总长
    let copied = sourceEnd - sourceStart

    if (copied > 0) {
      target.#replace(
        this.slice(sourceStart, sourceEnd),
        targetStart,
        targetStart + copied
      )
    }

    return copied
  }

  /**
   * 在指定位置, 写入指定内容
   */
  write(str = '', offset = 0, len, encoding = 'utf8') {
    len ??= this.length
    let written = len - offset
    let end = offset + written
    let u8 = any2u8(str, encoding)

    if (u8.length < written) {
      written = u8.length
      end = offset + written
    }
    if (written > 0) {
      this.#replace(u8, offset, end)
    }
    return written
  }

  /**
   * 与目标对象比较
   */
  compare(buf) {
    return compare(this, buf)
  }

  /**
   * 判断2个Buffer对象是否相等
   */
  equals(buf) {
    return this.toString() === buf.toString()
  }

  /**
   * 判断是否包含指定内容
   * 原生的includes只支持bytes, 这里拓展支持完整的类型
   */
  includes(data, offset = 0, encoding = 'utf8') {
    if (typeof data === 'number') {
      return super.includes(data, offset)
    }
    let u8 = any2u8(data, encoding)
    let str = decode(u8)

    return this.toString().indexOf(str, offset) > -1
  }
}
